import importlib.metadata

from groundhog_hpc.decorators import function, harness

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = ["function", "harness", "__version__"]



DEFAULT_USER_CONFIG = {
    "worker_init": "pip show -qq uv || pip install uv",  # install uv in the worker environment
}

DEFAULT_ENDPOINTS = {
    "anvil": "5aafb4c1-27b2-40d8-a038-a0277611868f",  # official anvil multi-user-endpoint
}

DEFAULT_WALLTIME_SEC = 60



import os
from pathlib import Path
from typing import Optional

import typer

import groundhog_hpc
from groundhog_hpc.errors import RemoteExecutionError

app = typer.Typer()


@app.command(no_args_is_help=True)
def run(
    script: Path = typer.Argument(
        ..., help="Python script with dependencies to deploy to the endpoint"
    ),
    function: str = typer.Argument(
        "main", help="Name of harness to run from script (default 'main')."
    ),
):
    """Run a Python script on a Globus Compute endpoint."""

    script_path = script.resolve()
    if not script_path.exists():
        typer.echo(f"Error: Script '{script_path}' not found", err=True)
        raise typer.Exit(1)
    else:
        # used by _Function to build callable
        os.environ["GROUNDHOG_SCRIPT_PATH"] = str(script_path)

    contents = script_path.read_text()

    try:
        # Use the same dict for both globals and locals so harness functions
        # can reference other top level functions
        script_namespace = {}

        exec(contents, script_namespace, script_namespace)

        if function not in script_namespace:
            typer.echo(f"Error: Function '{function}' not found in script", err=True)
            raise typer.Exit(1)

        result = script_namespace[function]()
        typer.echo(result)
    except RemoteExecutionError as e:
        typer.echo(f"Remote execution failed (exit code {e.returncode}):", err=True)
        typer.echo(e.stderr, err=True)
        raise typer.Exit(1)
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1)


def _version_callback(show):
    if show:
        typer.echo(f"{groundhog_hpc.__version__}")
    raise typer.Exit()


@app.callback(no_args_is_help=True)
def main_info(
    version: Optional[bool] = typer.Option(
        None, "--version", callback=_version_callback, is_eager=True
    ),
):
    """
    Hello, Groundhog ☀️🦫🕳️
    """
    pass



import functools
import os

from groundhog_hpc.function import Function
from groundhog_hpc.settings import DEFAULT_USER_CONFIG


def harness():
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            os.environ["GROUNDHOG_HARNESS"] = str(True)
            results = func(*args, **kwargs)
            del os.environ["GROUNDHOG_HARNESS"]
            return results

        return wrapper

    return decorator


def function(endpoint=None, walltime=None, **user_endpoint_config):
    if not user_endpoint_config:
        user_endpoint_config = DEFAULT_USER_CONFIG
    elif "worker_init" in user_endpoint_config:
        # ensure uv install command is part of worker init
        user_endpoint_config["worker_init"] += f"\n{DEFAULT_USER_CONFIG['worker_init']}"

    def decorator(func):
        wrapper = Function(func, endpoint, walltime, **user_endpoint_config)
        functools.update_wrapper(wrapper, func)
        return wrapper

    return decorator



import os
from typing import Callable

from groundhog_hpc.runner import script_to_callable
from groundhog_hpc.settings import DEFAULT_ENDPOINTS


class Function:
    def __init__(
        self,
        func: Callable,
        endpoint=None,
        walltime=None,
        **user_endpoint_config,
    ):
        self.script_path = os.environ.get("GROUNDHOG_SCRIPT_PATH")  # set by cli
        self.endpoint = endpoint or DEFAULT_ENDPOINTS["anvil"]
        self.walltime = walltime
        self.user_endpoint_config = user_endpoint_config

        self._local_func = func
        self._remote_func = None

    def __call__(self, *args, **kwargs):
        return self._local_func(*args, **kwargs)

    def remote(self, *args, **kwargs):
        if not self._running_in_harness():
            raise Exception(
                "Error: can't invoke a remote function outside of a @hog.harness function"
            )
        if self._remote_func is None:
            # delay defining the remote function until we're already invoking
            # the harness to avoid "No such file or directory: '<string>'" etc.
            # also avoids redefining the shell function recursively when running
            # on the remote endpoint
            self._remote_func = self._init_remote_func()

        return self._remote_func(*args, **kwargs)

    def _running_in_harness(self):
        # set by @harness decorator
        return bool(os.environ.get("GROUNDHOG_HARNESS"))

    def _init_remote_func(self):
        if self.script_path is None:
            raise ValueError("Could not locate source file")

        with open(self.script_path, "r") as f_in:
            self.contents = f_in.read()

        return script_to_callable(
            self.contents,
            self._local_func.__qualname__,
            self.endpoint,
            self.walltime,
            self.user_endpoint_config,
            self.script_path,
        )



# /// script
# requires-python = "==3.12"
# dependencies = ["torch"]
# ///
"""
sample script for `hog run` to execute on the remote globus compute endpoint with uv
"""

import sys

import torch


def main() -> None:
    major, minor, micro, *_ = sys.version_info
    version = f"{major}.{minor}.{micro}"
    print(f"Greetings from python {version} (with love)")


def hello_torch():
    print(f"{torch.cuda.is_available()=}")
    print(f"Greetings from torch {torch.__version__} (with love)")


if __name__ == "__main__":
    main()
    hello_torch()



"""Custom exceptions for Groundhog."""


class RemoteExecutionError(Exception):
    """Raised when a remote function execution fails on the Globus Compute endpoint.

    Attributes:
        message: Human-readable error description
        stderr: Standard error output from the remote execution
        returncode: Exit code from the remote process
    """

    def __init__(self, message: str, stderr: str, returncode: int):
        # Remove trailing WARNING lines that aren't part of the traceback
        lines = stderr.strip().split("\n")
        while lines and lines[-1].startswith("WARNING:"):
            lines.pop()

        self.stderr = "\n".join(lines)
        self.returncode = returncode
        super().__init__(message)



import warnings
from hashlib import sha1
from pathlib import Path
from typing import Callable
from uuid import UUID

from groundhog_hpc.errors import RemoteExecutionError
from groundhog_hpc.serialization import deserialize, serialize
from groundhog_hpc.settings import DEFAULT_USER_CONFIG

warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    module="globus_compute_sdk",
)

SHELL_COMMAND_TEMPLATE = """
cat > {script_basename}-{script_hash}.py << 'EOF'
{contents}
EOF
cat > {script_basename}-{script_hash}.in << 'END'
{payload}
END
$(python -c 'import uv; print(uv.find_uv_bin())') run --managed-python \\
  {script_basename}-{script_hash}.py {function_name} {script_basename}-{script_hash}.in > {script_basename}-{script_hash}-run.stdout \\
  && cat {script_basename}-{script_hash}.out
"""
# note: working directory is ~/.globus_compute/uep.<endpoint uuids>/tasks_working_dir


def script_to_callable(
    user_script: str,
    function_name: str,
    endpoint: str,
    walltime: int | None = None,
    user_endpoint_config: dict | None = None,
    script_path: str | None = None,
) -> Callable:
    """Create callable corresponding to the named function from a user's script.

    The created function accepts the same arguments as the original named function, but
    dispatches to a shell function on the remote endpoint.

    NOTE: The function must expect json-serializable input and return json-serializable output.
    """
    import globus_compute_sdk as gc  # lazy import so cryptography bindings don't break remote endpoint

    config = DEFAULT_USER_CONFIG.copy()
    config.update(user_endpoint_config or {})

    script_hash = _script_hash_prefix(user_script)
    script_basename = (
        _extract_script_basename(script_path) if script_path else "groundhog"
    )
    contents = _inject_script_boilerplate(
        user_script, function_name, script_hash, script_basename
    )

    def run(*args, **kwargs):
        shell_fn = gc.ShellFunction(cmd=SHELL_COMMAND_TEMPLATE, walltime=walltime)
        payload = serialize((args, kwargs))

        with gc.Executor(UUID(endpoint), user_endpoint_config=config) as executor:
            future = executor.submit(
                shell_fn,
                script_hash=script_hash,
                script_basename=script_basename,
                contents=contents,
                function_name=function_name,
                payload=payload,
            )

            shell_result: gc.ShellResult = future.result()

            if shell_result.returncode != 0:
                raise RemoteExecutionError(
                    message=f"Remote execution failed with exit code {shell_result.returncode}",
                    stderr=shell_result.stderr,
                    returncode=shell_result.returncode,
                )

            return deserialize(shell_result.stdout)

    return run


def _script_hash_prefix(contents: str, length=8) -> str:
    return str(sha1(bytes(contents, "utf-8")).hexdigest()[:length])


def _extract_script_basename(script_path: str) -> str:
    return Path(script_path).stem


def _inject_script_boilerplate(
    user_script: str, function_name: str, script_hash: str, script_basename: str
) -> str:
    assert "__main__" not in user_script, (
        "invalid user script: can't define custom `__main__` logic"
    )
    # TODO better validation errors
    # or see if we can use runpy to explicitly set __name__ (i.e. "__groundhog_main__")
    # TODO validate existence of PEP 723 script metadata
    #
    payload_path = f"{script_basename}-{script_hash}.in"
    outfile_path = f"{script_basename}-{script_hash}.out"

    script = f"""{user_script}
if __name__ == "__main__":
    import sys

    from groundhog_hpc.serialization import serialize, deserialize

    with open('{payload_path}', 'r') as f_in:
        payload = f_in.read()
        args, kwargs = deserialize(payload)

    results = {function_name}(*args, **kwargs)
    with open('{outfile_path}', 'w+') as f_out:
        contents = serialize(results)
        f_out.write(contents)
"""
    return script
